using System;
using System.Drawing;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Web;
using System.Xml;

namespace SwivelSecure.PINsafe
{
	/// <summary><c>PINsafeClient</c> provides conventient access to the
	///		authentication service of a PINsafe server</summary>
	public class PINsafeClient
	{
		// Server settings
		private PINsafeSettings settings;

		// Last request/response
		private string request;
		private string response;
		private XmlDocument responseDoc;
		
		// Last error
		private string error;

		// Last warnings
		private string[] warnings;

		public PINsafeClient()
		{
			settings = new PINsafeSettings();
		}

		public PINsafeClient(PINsafeSettings settings)
		{
			this.settings = settings;
		}

		/// <summary>Gets or sets the use of SSL for communication with the
		///		PINsafe server</summary>
		public bool SSL 
		{
			get {
				return settings.Ssl;
			}
		}

		/// <summary>Gets or sets the hostname or IP address of the PINsafe
		///		server.</summary>
		public string Server
		{
			get	{
				return settings.Server;
			}
		}

		/// <summary>Gets or sets the port used by the PINsafe server.</summary>
		public int Port
		{
			get {
				return settings.Port;
			}
		}

		/// <summary>Gets or sets the context used by the PINsafe server.</summary>
		public string Context
		{
			get	{
				return settings.Context;
			}
		}

		/// <summary>Gets or sets the secret used for communicating with the
		///		PINsafe server.</summary>
		public string Secret
		{
			get	{
				return settings.Secret;
			}
		}

		public bool AcceptSelfSigned
		{
			get { return settings.AcceptSelfSigned; }
		}

		/// <summary>The URL for accessing the PINsafe server.</summary>
		public string Address
		{
			get
			{
				StringBuilder addressBuilder = new StringBuilder();

				if (SSL)
					addressBuilder.Append("https://");
				else
					addressBuilder.Append("http://");

				addressBuilder.AppendFormat("{0}:{1}/{2}/AgentXML", Server, Port, Context);

				return addressBuilder.ToString();
			}
		}

		/// <summary>The URL for accessing the PINsafe server.</summary>
		private string ImageAddress
		{
			get
			{
				return GetServletAddress("SCImage");
			}
		}

		/// <summary>The URL for accessing the PINsafe server.</summary>
		private string GetServletAddress(String servlet)
		{
			StringBuilder addressBuilder = new StringBuilder();

			if (SSL)
				addressBuilder.Append("https://");
			else
				addressBuilder.Append("http://");

			addressBuilder.AppendFormat("{0}:{1}/{2}/{3}", Server, Port, Context, servlet);

			return addressBuilder.ToString();
		}

		/// <summary>The URL for accessing the PINsafe server.</summary>
		private string MessageAddress
		{
			get
			{
				return GetServletAddress("DCMessage");
			}
		}

		/// <summary>Gets the last error.</summary>
		public string Error
		{
			get	{
				return error;
			}
		}

		/// <summary>Gets the last warnings.</summary>
		public string[] Warnings
		{
			get 
			{
				return warnings;
			}
		}

		/// <summary>Gets the last XML request sent to the PINSafe server.</summary>
		public string Request
		{
			get {
				return request;
			}
		}

		/// <value>Gets the last XML response received from the PINsafe server.</value>
		public string Response
		{
			get	{
				return response;
			}
		}

		private string CookieSecret
		{
			get { return settings.CookieSecret; }
		}

		/// <summary>
		/// Check whether a user exists or not.
		/// </summary>
		/// <param name="username"></param>
		/// <returns></returns>
		public bool UserExists(string username)
		{
			try
			{
				// Create XML
				StringWriter buffer = new StringWriter();
				XmlTextWriter writer = new XmlTextWriter(buffer);

				writer.WriteStartDocument();
				writer.WriteStartElement("SASRequest");
				writer.WriteElementString("action", "exists");
				writer.WriteElementString("username", username);
				writer.WriteElementString("secret", Secret);
				writer.WriteEndElement();
				writer.Close();

				request = buffer.ToString();

				bool result = DoRequest();

				return result;
			}
			catch (Exception e)
			{
				error = e.Message;
				return false;
			}
		}

		/// <summary>Starts a new single channel session with the PINsafe 
		///		server.</summary>
		/// <param name="username">The username for the user.</param>
		/// <returns>Session ID for the started session, or an empty string if 
		///		an error occured.</returns>
		public string StartSession(string username)
		{
			string sessionID;

			try
			{
				// Create XML
				StringWriter buffer = new StringWriter();
				XmlTextWriter writer = new XmlTextWriter(buffer);

				writer.WriteStartDocument();
				writer.WriteStartElement("SASRequest");
				writer.WriteElementString("action", "sessionstart");
				writer.WriteElementString("username", username);
				writer.WriteElementString("secret", Secret);
				writer.WriteEndElement();
				writer.Close();

				request = buffer.ToString();

				if (DoRequest()) {
					// Parse out session ID
					XmlNodeList sessionNodes = responseDoc.GetElementsByTagName("SessionID");
					if (sessionNodes.Count > 0)	{
						sessionID = sessionNodes[0].InnerText;
					} else {
						sessionID = "";
					}
				} else {
					sessionID = "";
				}
			}
			catch (Exception e)
			{
				error = e.Message;
				sessionID = "";
			}

			return sessionID;
		}

		/// <summary>Perform an authentication with the PINsafe server.</summary>
		/// <param name="username">The username of the user.</param>
		/// <param name="password">The user's password.</param>
		/// <param name="OTC">The user's one time code.</param>
		/// <returns>True if the authentication was successful, otherwise 
		///		false.</returns>
		public bool Login(string username, string password, string OTC)
		{
			bool result;

			try
			{
				// Create XML
				StringWriter buffer = new StringWriter();
				XmlTextWriter writer = new XmlTextWriter(buffer);

				writer.WriteStartDocument();
				writer.WriteStartElement("SASRequest");
				writer.WriteElementString("action", "login");
				writer.WriteElementString("username", username);
				writer.WriteElementString("password", password);
				writer.WriteElementString("otc", OTC);
				writer.WriteElementString("secret", Secret);
				writer.WriteEndElement();
				writer.Close();

				request = buffer.ToString();

				// Attempt authentication
				result = DoRequest();
			}
			catch (Exception e)
			{
				error = e.Message;
				result = false;
			}

			return result;
		}

		/// <summary>Perform a PIN change with the PINsafe server.</summary>
		/// <param name="username">The username of the user.</param>
		/// <param name="password">The user's current password.</param>
		/// <param name="newPassword">The user's new password.</param>
		/// <param name="OTC">The one time code for the user's current 
		///		PIN.</param>
		/// <param name="newOTC">The one time code for the user's new 
		///		PIN.</param>
		/// <returns>True if the PIN change was successful, otherwise 
		///		false</returns>
		public bool ChangePin(string username, string password, string newPassword,
							  string OTC, string newOTC)
		{
			bool result; 

			try {
				// Create XML
				StringWriter buffer = new StringWriter();
				XmlTextWriter writer = new XmlTextWriter(buffer);

				writer.WriteStartDocument();
				writer.WriteStartElement("SASRequest");
				writer.WriteElementString("action", "changepin");
				writer.WriteElementString("username", username);
				writer.WriteElementString("password", password);
				writer.WriteElementString("newpassword", newPassword);
				writer.WriteElementString("otc", OTC);
				writer.WriteElementString("newotc", newOTC);
				writer.WriteElementString("secret", Secret);
				writer.WriteEndElement();
				writer.Close();

				request = buffer.ToString();

				// Attempt pin change
				result = DoRequest();
			}
			catch (Exception e)
			{
				error = e.Message;
				result = false;
			}

			return result;
		}

		/// <summary>Send an XML request to the PINsafe server.</summary>
		/// <param name="xml">The XML request.</param>
		/// <returns>The XML response from the server, or an empty string if 
		///		an error occured</returns>
		public string SendRequestXml(String xml)
		{
			request = xml;
			SendRequest();
			return response;
		}

		/// <summary>Send the request to the PINsafe server and parse the result.</summary>
		/// <returns>True if the request was successful, false otherwise.</returns>
		private bool DoRequest()
		{
			bool result;

			SendRequest();

			// Parse response
			responseDoc = new XmlDocument();
			responseDoc.LoadXml(response);
			XmlNodeList resultNodes = responseDoc.GetElementsByTagName("Result");

			if ((resultNodes.Count == 1) && resultNodes[0].InnerText.Equals("PASS")) {
				result = true;

				// Parse out any warnings
				XmlNodeList warningNodes = responseDoc.GetElementsByTagName("Warning");
				warnings = new string[warningNodes.Count];
				for (int i=0; i<warningNodes.Count; i++) 
				{
					warnings[i] = warningNodes[i].InnerText;
				}
			} else {
				result = false;

				// Parse out the error
				XmlNodeList errorNodes = responseDoc.GetElementsByTagName("Error");
				if (errorNodes.Count > 0) {
					error = errorNodes[0].InnerText;
				}
			}

			return result;
		}

		/// <summary>Send the request to the PINsafe server.</summary>
		private void SendRequest()
		{
			response = "";
			
			byte[] requestData = Encoding.UTF8.GetBytes(HttpUtility.UrlEncode(request));

			WebClient client = new WebClient();
			if (AcceptSelfSigned)
			{
				ServicePointManager.ServerCertificateValidationCallback = DummyCertificateValidator;
			}
			byte[] responseData = client.UploadData(Address, requestData);

			response = Encoding.UTF8.GetString(responseData);
		}

		private bool DummyCertificateValidator(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
		{
			return true;
		}

		private static Image BlankImage
		{
			get
			{
				Bitmap img = new Bitmap(1, 1);
				img.SetPixel(0, 0, Color.White);
				return img;
			}
		}

		private Image GetImage(String sessionID, String username, string baseAddress, out int status)
		{
			if (sessionID == null || sessionID == "")
			{
				if (username != null && username != "")
				{
					sessionID = StartSession(username);
				}
				else
				{
					status = (int)HttpStatusCode.OK;
					return BlankImage;
				}
			}

			if (sessionID == null && sessionID == "")
			{
				status = (int)HttpStatusCode.OK;
				return BlankImage;
			}

			StringBuilder address = new StringBuilder(baseAddress);
			address.Append("?sessionid=");
			address.Append(sessionID);
			address.Append("&random=");
			address.Append(new Random().Next());
			string contentType;
			HttpStatusCode statusCode;
			Image image = GetPinsafeResponse(address.ToString(), out contentType, out statusCode);
			status = (int)statusCode;
			if (statusCode != HttpStatusCode.OK)
			{
				return BlankImage;
			}
			return image;
		}

		public Image GetImage(String sessionID, String username, out int status)
		{
			return GetImage(sessionID, username, ImageAddress, out status);
		}

		public Image GetMessage(String sessionID, String username, out int status)
		{
			return GetImage(sessionID, username, MessageAddress, out status);
		}

		/// <summary>
		/// Send a request to the PINSafe server, and get the image data.
		/// </summary>
		/// <param name="requestUrl">The URL of the PINSafe request.</param>
		/// <param name="contentType">The content type of the returned image.</param>
		/// <param name="status">The response code from the PINsafe server.</param>
		/// <returns>The PINSafe image data.</returns>
		private static Image GetPinsafeResponse(string requestUrl, out string contentType, out HttpStatusCode status)
		{
			HttpWebRequest request = (HttpWebRequest)WebRequest.Create(requestUrl);
			HttpWebResponse response = (HttpWebResponse)request.GetResponse();
			status = response.StatusCode;
			if (status == HttpStatusCode.OK)
			{
				MemoryStream strm = new MemoryStream();
				Stream respStream = response.GetResponseStream();
				int bufSize = 0x400;
				byte[] buffer = new byte[bufSize];
				int bufLen = respStream.Read(buffer, 0, bufSize);
				while (bufLen > 0)
				{
					strm.Write(buffer, 0, bufLen);
					bufLen = respStream.Read(buffer, 0, bufSize);
				}
				contentType = response.ContentType;
				return Image.FromStream(strm);
			}
			contentType = "";
			return null;
		}

		public string LoginCookie(string username, string password, string otc, DateTime expiryTime)
		{
			return Login(username, password, otc) ? GetCookie(username, expiryTime) : null;
		}

		private string GetCookie(string username, DateTime expiryTime)
		{
			TimeSpan expirySpan = expiryTime.Subtract(new DateTime(1970, 1, 1));
			string expirySecs = ((long)expirySpan.TotalSeconds).ToString("D10");
			string timeAndUser = expirySecs + username;
			return GetHashValue(timeAndUser) + timeAndUser;
		}

		public string UpdateCookie(string oldCookie)
		{
			if (CookieValid(oldCookie))
			{
				string sExpiryTime = oldCookie.Substring(32, 10);
				DateTime oldExpiryTime = new DateTime(1970, 1, 1).AddSeconds(long.Parse(sExpiryTime));
				TimeSpan timeToExpire = oldExpiryTime - DateTime.Now;
				if (timeToExpire.TotalSeconds < settings.IdleTimeSecs - 10)
				{
					// Cookie is more than 10 seconds old - renew it.
					string username = oldCookie.Substring(42);
					DateTime expiryTime = DateTime.Now.AddSeconds(settings.IdleTimeSecs);
					return GetCookie(username, expiryTime);
				}
			}
			return null;
		}

		private string GetHashValue(string timeAndUser)
		{
			string data = CookieSecret + timeAndUser;
			byte[] md5 = MD5.Create().ComputeHash(Encoding.UTF8.GetBytes(data));
			StringBuilder hash = new StringBuilder();
			foreach (byte b in md5)
			{
				hash.Append(b.ToString("X2"));
			}
			return hash.ToString();
		}

		public bool CookieValid(string cookieData)
		{
			// Split the cookie up.
			if (cookieData == null || cookieData.Length <= 42)
			{
				return false;
			}
			string expiryString = cookieData.Substring(32, 10);
			long expiryLong;
			if (!long.TryParse(expiryString, out expiryLong))
			{
				return false;
			}
			DateTime expiryTime = new DateTime(1970, 1, 1).AddSeconds(expiryLong);
			if (expiryTime < DateTime.Now)
			{
				return false; // Cookie has expired.
			}
			string orgHash = cookieData.Substring(0, 32);
			string username = cookieData.Substring(42);
			return (orgHash.CompareTo(GetHashValue(expiryString + username)) == 0);
		}
	}
}
